import collections
from . import constants


control_codes = collections.OrderedDict()
control_codes[bytes([0xff, 17])] = b'\\himself'
control_codes[bytes([0xff, 1])] = b'\\1'
control_codes[bytes([0xff, 2])] = b'\\2'
control_codes[bytes([0xff, 3])] = b'\\3'
control_codes[bytes([0xff, 5])] = b'\\th'
control_codes[bytes([0xff, 6])] = b'\\a'
control_codes[bytes([0xff, 7])] = b'\\A'
control_codes[bytes([0xff, 8])] = b'\\the'
control_codes[bytes([0xff, 9])] = b'\\The'
control_codes[bytes([0xff, 10])] = b'\\he'
control_codes[bytes([0xff, 11])] = b'\\He'
control_codes[bytes([0xff, 12])] = b'\\his'
control_codes[bytes([0xff, 13])] = b'\\His'
control_codes[bytes([0xff, 14])] = b'\\hers'
control_codes[bytes([0xff, 15])] = b'\\Hers'
control_codes[bytes([0xff, 16])] = b'\\him'
control_codes[bytes([0xff, 18])] = b'\\...'
control_codes[bytes([0xff, 20])] = b'\\s'
control_codes[bytes([0xff, 21])] = b'\\proper'
control_codes[bytes([0xff, 22])] = b'\\improper'
control_codes[bytes([0xff, 23])] = b'\\bold'
control_codes[bytes([0xff, 24])] = b'\\italic'
control_codes[bytes([0xff, 25])] = b'\\underline'
control_codes[bytes([0xff, 27])] = b'\\font'
control_codes[bytes([0xff, 28])] = b'\\color'
control_codes[bytes([0xff, 30])] = b'\\resetcolor'  # TODO
control_codes[bytes([0xff, 31])] = b'\\red'
control_codes[bytes([0xff, 32])] = b'\\green'
control_codes[bytes([0xff, 33])] = b'\\blue'
control_codes[bytes([0xff, 34])] = b'\\black'
control_codes[bytes([0xff, 35])] = b'\\white'
control_codes[bytes([0xff, 36])] = b'\\yellow'
control_codes[bytes([0xff, 37])] = b'\\cyan'
control_codes[bytes([0xff, 38])] = b'\\magenta'
control_codes[bytes([0xff, 39])] = b'\\beep'
control_codes[bytes([0xff, 40])] = b'\\link'
control_codes[bytes([0xff, 42])] = b'\\ref'
control_codes[bytes([0xff, 43])] = b'\\icon'
control_codes[bytes([0xff, 44])] = b'\\Roman'
control_codes[bytes([0xff, 45])] = b'\\Roman'  # reset code


class RawString:
    def __init__(self, byte, key, mode=constants.raw_string_mode_encrypted, lazy=False):
        self.key = key
        self.orig_key = key
        if mode == constants.raw_string_mode_encrypted:
            self.orig_data = byte
            self.data = None
            self.string = None
            if not lazy:
                self.decrypt(True)
                self.decode(True)
        elif mode == constants.raw_string_mode_decrypted:
            self.data = byte
            self.string = None
            self.orig_data = None
            if not lazy:
                self.encrypt(True)
                self.decode(True)
        elif mode == constants.raw_string_mode_string:
            self.string = byte
            self.data = None
            self.orig_data = None
            if not lazy:
                self.encrypt(True)
                self.encode(True)

    # orig_data (encrypted bytes) -> data (decrypted bytes)
    def decrypt(self, force=False):
        if not force and self.data is not None:
            return self.data
        self.data = bytearray(self.orig_data)
        for i in range(len(self.orig_data)):
            self.data[i] = (self.orig_data[i] ^ self.key) & 0xFF
            self.key += 9
        return self.data

    # data (decrypted bytes) -> orig_data (encrypted bytes)
    def encrypt(self, force=False):
        if self.data is None:
            self.encode()
        if not force and self.orig_data is not None:
            return self.orig_data
        self.orig_data = bytearray(self.data)
        for i in range(len(self.data)):
            self.orig_data[i] = (self.data[i] ^ self.key) & 0xFF
            self.key += 9
        return self.orig_data

    # data (decrypted bytes) -> string (human-readable string)
    def decode(self, force=False):
        if self.data is None:
            self.decrypt()
        if self.string is None or force:
            temp = bytearray(self.data)
            for code, replacement in control_codes.items():
                temp = temp.replace(code, replacement)
            ccindex = temp.find(bytes([0xff]))
            if ccindex != -1:
                if len(temp) > ccindex + 1:
                    raise ValueError("Found unhandled control code: {0}".format(self.data[ccindex + 1]))
                else:
                    return "?"
            self.string = temp.decode('iso-8859-1')
            self.key = self.orig_key
        return self.string

    # string (human-readable string) -> data (decrypted bytes)
    def encode(self, force=False):
        if self.data is None or force:
            self.data = self.string.encode('iso-8859-1')
            for code, replacement in control_codes.items():
                self.data = self.data.replace(replacement, code)
            self.key = self.orig_key
        return self.data

    def __bytes__(self):
        return self.decrypt()

    def __str__(self):
        return self.decode()

    def __repr__(self):
        return "RawString(b'{0}',{1})".format(self.orig_data, self.orig_key)


class Mob:
    def __init__(self):
        self._unknown = 0
        self._fdata1 = 0
        self._fdata2 = 0


class Resource:
    def __init__(self, typeid, rhash):
        self.typeid = typeid
        self.hash = rhash


astype_map = {
    1: "mob",
    2: "obj",
    4: "text",
    8: "num",
    16: "file",
    32: "turf",
    128: "null",
    256: "area",
    1024: "sound",
    2048: "message",
    4096: "anything",
}


def resolve_astype(flags):
    types = []
    for pos in astype_map:
        if flags & pos > 0:
            types.append(astype_map[pos])
    return types


intype_map = {
    1: "view",
    2: "oview",
    3: "usr.loc",
    8: "usr",
    16: "world",
    64: "custom",
}


intype_has_parameter = ["view", "oview"]


class Arg:
    def __init__(self, name, astype=0, intype=1, inrange=125):
        self.name = name
        if isinstance(astype, list):
            self.astype = astype
        else:
            self.astype = resolve_astype(astype)
        if isinstance(intype, int):
            self.intype = intype_map[intype]
        else:
            self.intype = intype
        if not isinstance(inrange, Proc) and inrange == 125:
            self.inrange = None
        else:
            self.inrange = inrange

    def __str__(self):
        data = self.name
        if len(self.astype):
            data += " as {0}".format("|".join(self.astype))
        if self.intype is not None:
            if self.intype != "custom":
                if self.intype in intype_has_parameter:
                    if self.inrange is not None:
                        data += " in {0}({1})".format(self.intype, self.inrange)
                    else:
                        data += " in {0}()".format(self.intype)
                else:
                    data += " in {0}".format(self.intype)
            else:
                data += " in <proc:{0}>".format(self.inrange.argproc_id)

    def __repr__(self):
        return "Arg({0}, astype={1}, intype={2}, inrange={3})".format(repr(self.name), repr(self.astype), repr(self.intype), repr(self.inrange))


class Proc:
    def __init__(self):
        self.path = 0
        self.name = 0
        self.desc = 0
        self.category = 0
        self.data = 0
        self.variable_list = 0
        self.argument_list = 0

        self.range = 0
        self.access = 0
        self.flags = 0
        # 0x80: use extended flags
        # 0x20: popup_menu
        #
        # 0x01: hidden
        self._ufield3 = 0
        self.ext_flags = 0
        self.invisibility = 0

        self.parameters = []
        self.argproc_id = None
        self.locals = []
        self.defined_on = None


class Var:
    def __init__(self):
        self.value = None,
        self.name = 0


class Instance:
    def __init__(self):
        self.value = None,
        self.initializer = 0


class WorldData:
    def __init__(self):
        self.world_version = 0
        self.min_server = 0
        self.min_client = 0
        self.map_x = 0
        self.map_y = 0
        self.map_z = 0

        self.default_mob = 0
        self.default_turf = 0
        self.default_area = 0
        self.world_procs = 0
        self.global_init = 0
        self.world_domain = 0
        self.world_name = 0
        self.tick_lag = 0
        self.unknown1 = 0
        self.client_type = 0
        self.image_type = 0
        self.lazy_eye = 0
        self.client_dir = 0
        self.control_freak = 0
        self.unknown2 = 0
        self.client_script = 0
        self.unknown3 = 0
        self.view_width = 0
        self.view_height = 0
        self.hub_password = 0
        self.world_status = 0
        self.unknown4 = 0
        self.unknown5 = 0
        self.version = 0
        self.cache_lifespan = 0
        self.default_command_text = 0
        self.default_command_prompt = 0
        self.hub_path = 0
        self.unknown6 = 0
        self.unknown7 = 0
        self.icon_width = 0
        self.icon_height = 0
        self.map_format = 0


class Tile:
    def __init__(self, area_id=0, turf_id=0, unknown_id=0):
        self.area = area_id
        self.turf = turf_id
        self.unknown = unknown_id
        self.instances = []


class Type:
    def __init__(self, path, parent):
        self.path = path
        self.parent = parent

        self.name = 0
        self.desc = 0
        self.icon = 0
        self.icon_state = 0
        self.dir = 0

        self.text = 0
        self.suffix = 0

        self.maptext = 0
        self.maptext_width = 0
        self.maptext_height = 0
        self.maptext_x = 0
        self.maptext_y = 0

        self.flags = 0
        self.proc_list = 0
        self.verb_list = 0
        self.variable_list = 0
        self.layer = 0.0
        self.builtin_variable_list = 0

        self.id = 0
        self.resolved_vars = False
        self.variables = None
        self.procedures = None
        self.verbs = None
        self.procedures_own = {}
        self.verbs_own = {}

        self._unknown1 = 0
        self._unknown2 = 0
        self._unknown3 = 0
        self._unknown4 = 0
        self._fdata1 = 0
        self._fdata4 = 0

    def __json__(self):
        return {
            "path": self.path,
            "name": self.name,
            "desc": self.desc,
            "parent": self.parent.path,
        }

    def __str__(self):
        return '"Type: {0}"'.format(self.path)

    def __repr__(self):
        return '"{0}"'.format(self.path)
        try:
            return 'type({0}, {1})'.format(self.path, self.parent.path)
        except:
            return 'type({0}, {1})'.format(self.path, self.parent)



class compiled_value():
    def __init__(self, typeid, value):
        self._typeid = typeid
        self._value = value


class value_null(compiled_value):
    pass


class value_mob(compiled_value):
    def __init__(self, typeid, value):
        super().__init__(typeid, value)
        self.value = value


class value_resource(compiled_value):
    def __init__(self, typeid, value):
        super().__init__(typeid, value)
        self.value = value


class value_type(compiled_value):
    def __init__(self, typeid, value):
        super().__init__(typeid, value)
        self.value = value


class value_savefile_type(compiled_value):
    def __init__(self, typeid, value):
        super().__init__(typeid, value)
        pass


class value_file_type(compiled_value):
    def __init__(self, typeid, value):
        super().__init__(typeid, value)
        pass


class value_list_type(compiled_value):
    def __init__(self, typeid, value):
        super().__init__(typeid, value)
        pass


class value_client_type(compiled_value):
    def __init__(self, typeid, value):
        super().__init__(typeid, value)
        pass


class value_string(compiled_value):
    def __init__(self, typeid, value):
        super().__init__(typeid, value)
        self.value = value


class value_number(compiled_value):
    def __init__(self, typeid, value):
        super().__init__(typeid, value)
        self.value = value


class value_list(compiled_value):
    def __init__(self, typeid, value):
        super().__init__(typeid, value)
        self.value = value


class value_proc(compiled_value):
    def __init__(self, typeid, value):
        super().__init__(typeid, value)
        self.value = value


class value_image(compiled_value):
    def __init__(self, typeid, value):
        super().__init__(typeid, value)
        self.value = value


class value_unknown(compiled_value):
    def __init__(self, typeid, value):
        super().__init__(typeid, value)
        self.typeid = typeid
        self.value = value


decode_map = {
    0: value_null,
    6: value_string,
    8: value_mob,
    9: value_type,
    10: value_type,
    11: value_type,
    12: value_resource,
    32: value_type,
    36: value_savefile_type,
    39: value_file_type,
    40: value_list_type,
    42: value_number,
    59: value_client_type,
    62: value_list,
    63: value_image,
}


def decode(typeid, value):
    global decode_map
    if typeid in decode_map:
        return decode_map[typeid](typeid, value)
    return value_unknown(typeid, value)



from .dmb import Type, Proc, Arg, resolve_astype
from . import json as dmbjson
from . import constants
import json
import struct


def tree_from_json(jdict):
    tree = ObjectTree(tree=jdict)
    tree.resolve_parent_refs_recursive(tree.tree)
    return tree

typemap = {
    0: "null",
    6: "string",
    8: "mob",
    9: "type",
    10: "type",
    11: "type",
    12: "resource",
    32: "type",
    36: "savefile",
    39: "file",
    40: "list",
    42: "float",
    59: "client",
    62: "list",
    63: "image",
}

skips = {
    "null": 1,
    "string": 1,
    "float": 2,
}

datum_vars = set(["parent_type", "tag", "type", "vars"])
atom_vars = set(["alpha", "appearance", "appearance_flags", "blend_mode", "color", "density", "desc", "dir", "gender", "icon", "icon_state", "infra_luminosity", "invisibility", "layer", "loc", "maptext", "maptext_width", "maptext_height", "maptext_x", "maptext_y", "mouse_drag_pointer", "mouse_drop_pointer", "mouse_drop_zone", "mouse_opacity", "mouse_over_pointer", "name", "opacity", "overlays", "override", "pixel_x", "pixel_y", "pixel_z", "plane", "suffix", "text", "transform", "underlays", "verbs", "x", "y", "z"]) | datum_vars
movable_vars = set(["animate_movement", "bound_x", "bound_y", "bound_width", "bound_height", "locs", "glide_size", "screen_loc", "step_size", "step_x", "step_y"]) | atom_vars
base_type_variables = {
    "area": atom_vars,
    "atom": atom_vars,
    "datum": datum_vars,
    "image": set(["loc"]),
    "list": set(["len"]),
    "mob": movable_vars | set(["ckey", "client", "group", "key", "see_in_dark", "see_infrared", "see_invisible", "sight"]),
    "obj": movable_vars,
    "savefile": set(["cd", "dir", "eof", "name"]),
    "sound": set(["file", "repeat", "wait", "channel", "volume", "frequency", "pan", "priority", "status", "x", "y", "z", "falloff", "environment", "echo"]),
    "turf": atom_vars,
    "world": set(["address", "area", "byond_version", "cache_lifespan", "contents", "cpu", "executor", "fps", "game_state", "host", "hub", "hub_password", "icon_size", "internet_address", "log", "loop_checks", "map_format", "maxx", "maxy", "maxz", "mob", "name", "params", "port", "realtime", "reachable", "sleep_offline", "status", "system_type", "tick_lag", "time", "timeofday", "turf", "url", "version", "view", "visibility"]),
}


def base_type_vars(path):
    btv = path.lstrip("/")
    if btv in base_type_variables:
        return set() | base_type_variables[btv]
    return set()


class ObjectTree:
    def __init__(self, tree={}, dmb=None, loaded=False):
        if dmb is not None and dmb.string_mode == constants.string_mode_byte_strings:
            raise TypeError("ObjectTrees cannot be used with undecoded strings.")
        self.tree = tree

        if loaded and tree != {}:
            self._resolve_tree_recursive(self.tree)
        if dmb is not None:
            self._populate_types_from_dmb(dmb)
        if loaded or dmb is not None:
            self._resolve_parents_recursive(self.tree)
        if dmb is not None:
            self._populate_variables_recursive(self.tree, dmb)
            self._assign_procedures(dmb)

    def _populate_types_from_dmb(self, dmb):
        for t in dmb.types:
            pt = t.parent
            if pt == 65535:
                parent = dmb.no_parent_type
            else:
                parent = dmb._resolve_string(dmb.types[pt].path)
            res = Type(dmb._resolve_string(t.path), parent)
            res.name = dmb._resolve_string(t.name)
            res.desc = dmb._resolve_string(t.desc)
            res.icon = dmb._resolve_resource(res.icon)
            res.icon_state = dmb._resolve_string(t.icon_state)
            res.variable_list = t.variable_list
            res.builtin_variable_list = t.builtin_variable_list
            res.proc_list = t.proc_list
            res.verb_list = t.verb_list

            res.flags = t.flags
            res.dir = t.dir
            res.layer = t.layer
            res.id = t.id
            self.push(res)

    def _resolve_and_collect_vars_recursive(self, t, dmb):
        if t.variables is not None:
            return
        varset = base_type_vars(t.path)
        p = t.parent

        varlist = dmb._resolve_data(t.variable_list)
        if varlist is not None:
            databytes = dmb._unpack_arch(varlist)
            skip = False
            for varid in databytes:
                if skip:
                    skip = False
                    continue
                var = dmb._resolve_var(varid)
                varset.add(dmb._resolve_string(var.name))
                skip = True
        bvarlist = dmb._resolve_data(t.builtin_variable_list)
        if bvarlist is not None:
            databytes = dmb._unpack_arch(bvarlist)
        #     print(t.path)
        #     print(t.builtin_variable_list, "( =", bvarlist, ")")
        #     print(databytes)
            i = 0
            while i < len(databytes):
                # varname = databytes[i]
                vartype = databytes[i+1]
                vartype_str = typemap[vartype]
                if vartype_str not in skips:
                    print(t.builtin_variable_list, "( =", bvarlist, ")")
                    print(databytes)
                    raise ValueError(vartype_str)
                i += 2 + skips[vartype_str]
        #         varset.add(varname)
        # #     skip = False
        # #     for strid in databytes:
        # #         if skip:
        # #             skip = False
        # #             continue
        # #         var = dmb._resolve_string(strid)
        # #         print(strid, "( =", var, ")")
        # #         varset.add(var)
        # #         skip = True
        if p is not dmb.no_parent_type:
            if p.variables is None:
                self._resolve_and_collect_vars_recursive(p, dmb)
            varset |= p.variables
        t.variables = varset

    def _populate_variables_recursive(self, currtree, dmb):
        for k in currtree:
            if k == '.':
                self._resolve_and_collect_vars_recursive(currtree[k], dmb)
            else:
                self._populate_variables_recursive(currtree[k], dmb)

    def _unpack_proc_locals(self, proc, dmb):
        proc.locals = []
        vlist = dmb._resolve_data(proc.variable_list)
        if vlist is not None:
            vs = dmb._unpack_arch(vlist)
            if len(vs) > 0:
                for v in vs:
                    proc.locals.append(dmb._resolve_string(dmb._resolve_var(v).name))

    def _assign_procedures(self, dmb):
        for p in dmb.procs:
            if p.path == 65535:
                continue
            c = Proc()
            c.path = dmb._resolve_string(p.path)
            c.name = dmb._resolve_string(p.name)
            c.desc = dmb._resolve_string(p.desc)
            c.category = dmb._resolve_string(p.category)
            c.range = p.range
            c.access = p.access
            c.flags = p.flags
            c.ext_flags = p.ext_flags
            c.invisibility = p.invisibility
            c.variable_list = p.variable_list
            self._unpack_proc_locals(c, dmb)

            alist = dmb._resolve_data(p.argument_list)
            if alist is not None:
                args = list(dmb._unpack_arch(alist))
                if len(args) > 0:
                    i = 0
                    while i < len(args):
                        args[i+1] = struct.unpack("<BBBB", struct.pack("<I", args[i+1]))
                        v = dmb._resolve_var(args[i+2])
                        argname = dmb._resolve_string(v.name)
                        inrange = args[i+1][1]
                        if args[i+1][0] == 64:
                            argproc = dmb.argprocs[args[i+1][1]]
                            argproc_proc = dmb._resolve_proc(argproc)
                            inrange = Proc()
                            inrange.data = argproc_proc.data
                            inrange.argproc_id = args[i+1][1]
                            inrange.variable_list = argproc_proc.variable_list
                            self._unpack_proc_locals(inrange, dmb)
                        c.parameters.append(Arg(argname, args[i], args[i+1][0], inrange))
                        i += 4

            parts = c.path.rsplit("/", 2)
            is_verb = False
            if parts[1] != "proc":
                if parts[1] != "verb":
                    parts[0] += "/{0}".format(parts[1])
                else:
                    is_verb = True
            stree = self.get_path(parts[0])
            if stree is not None:
                t = stree["."]
                if not is_verb:
                    t.procedures_own[parts[2]] = c
                else:
                    t.verbs_own[parts[2]] = c
                c.defined_on = t

        self._populate_procedures_recursive(self.tree, dmb)

    def _collect_procedures(self, t, dmb):
        t.procedures = {}
        t.verbs = {}
        for pname in t.procedures_own:
            t.procedures[pname] = t.procedures_own[pname]
        for vname in t.verbs_own:
            t.verbs[pname] = t.verbs_own[vname]
        p = t.parent
        if p is dmb.no_parent_type:
            return
        if p.procedures is None:
            self._collect_procedures(p, dmb)
        if p.procedures is not None:
            for pname in p.procedures:
                if pname not in t.procedures:
                    t.procedures[pname] = p.procedures[pname]
        if p.verbs is not None:
            for vname in p.verbs:
                if vname not in t.verbs:
                    t.verbs[pname] = p.verbs[vname]

    def _populate_procedures_recursive(self, currtree, dmb):
        for k in currtree:
            if k == '.':
                self._collect_procedures(currtree[k], dmb)
            else:
                self._populate_procedures_recursive(currtree[k], dmb)

    def _resolve_tree_recursive(self, currtree):
        for k in currtree:
            if k == '.' and isinstance(currtree[k], dict):
                old = currtree[k]
                currtree[k] = Type(old['path'], old['parent'])
                currtree[k].name = old['name']
                currtree[k].desc = old['desc']
                currtree[k].icon = old['icon']
                currtree[k].icon_state = old['icon_state']
                currtree[k].dir = old['dir']
                currtree[k].layer = old['layer']
                currtree[k].id = old['id']
            else:
                self._resolve_tree_recursive(currtree[k])

    def _resolve_parents_recursive(self, currtree):
        for k in currtree:
            if k == '.':
                p = currtree[k].parent
                if isinstance(p, str):
                    parent_tree = self.get_path(p)
                    if parent_tree is not None:
                        currtree[k].parent = parent_tree['.']
            else:
                self._resolve_parents_recursive(currtree[k])

    def __getitem__(self, key):
        return self.tree[key]

    def __iter__(self):
        return self.tree.__iter__()

    def push(self, typeinst):
        path = typeinst.path.split('/')
        curr_tree = self.tree
        for part in path[1:]:
            if part not in curr_tree:
                curr_tree[part] = {}
            curr_tree = curr_tree[part]
        curr_tree["."] = typeinst

    def json(self):
        return json.dumps(self.tree, cls=dmbjson.JSONEncoder, sort_keys=True, indent=2)

    def get_path(self, path):
        slash = b'/'
        if isinstance(path, str):
            slash = "/"
        path = path.lstrip(slash)
        parts = path.split(slash)
        curr_tree = self.tree
        for part in parts:
            if part in curr_tree:
                curr_tree = curr_tree[part]
            else:
                return None
        return ObjectTree(tree=curr_tree)

    def complete_path(self, path):
        parts = path.split('/')
        tree = self.get_path('/'.join(parts[:-1]))
        if tree is None:
            return None
        options = []
        for k in tree:
            if k == ".":
                continue
            if parts[-1] in k:
                options.append(k)
        return {
            "tree": tree,
            "options": options,
        }

    def __repr__(self):
        return self.json()



import struct
from . import constants
from .dmb import RawString, Type
from .crypt import byond32
import numpy as np
import io


class DmbWriter:
    def __init__(self, dmbname, dmb):
        self.dmb = dmb
        self.writer = open(dmbname, 'wb')

    def __del__(self):
        if self.writer is not None:
            self.writer.close()
            self.writer = None

    def _uint8(self, v):
        data = struct.pack('<B', v & 0xFF)    # interprets the provided integer as 1 byte
        self.writer.write(data)
        return 1

    def _uint16(self, v):
        data = struct.pack('<H', v & 0xFFFF)    # interprets the provided integer as 2 LE bytes
        self.writer.write(data)
        return 2

    def _uint32(self, v):
        data = struct.pack('<I', v & 0xFFFFFFFF)    # interprets the provided integer as 4 LE bytes
        self.writer.write(data)
        return 4

    def _float32(self, v):
        data = struct.pack('<f', v)
        self.writer.write(data)
        return 4

    def _uarch(self, v):
        if self.dmb.bit32:
            return self._uint32(v)
        else:
            return self._uint16(v)

    def _nbytes(self, count, bs):
        if isinstance(bs, str):
            bs = bs.encode('iso-8859-1')
        elif isinstance(bs, list):
            bs = bytes(bs)
        avail = len(bs)
        if count > avail:
            raise IndexError("Cannot write {0} bytes, provided object only contains {1} bytes.".format(count, avail))
        elif count == avail:
            self.writer.write(bs)
        else:
            self.writer.write(bs[0:count])
        return count

    def _bytes(self, bs):
        self.writer.write(bs)
        return len(bs)

    def _str(self, data):
        e = data.encode('iso-8859-1')
        self.writer.write(e)
        return len(e)

    def _byte(self, bs):
        if isinstance(bs, int):
            self._nbytes(1, [bs])
        else:
            self._nbytes(1, bs)

    def _write_version_data(self):
        s1 = "world bin v{0}".format(self.dmb.world.world_version)
        self._str(s1)
        self._byte(b'\x0A')
        s2 = "min compatibility v{0} {1}".format(self.dmb.world.min_server, self.dmb.world.min_client)
        self._str(s2)
        self._byte(b'\x0A')
        return len(s1) + len(s2) + 2

    def _write_tiles(self):
        length = self.dmb.world.map_x * self.dmb.world.map_y * self.dmb.world.map_z
        x = 0
        y = 0
        z = 0
        rle = 0
        last_match = (None, None, None)
        lxr = len(self.dmb.tiles[z][y])
        lyr = len(self.dmb.tiles[z])
        lzr = len(self.dmb.tiles)
        while length > 0:
            tile = self.dmb.tiles[z][y][x]
            if isinstance(tile.area, Type):
                area = tile.area.id
            else:
                area = tile.area
            if isinstance(tile.turf, Type):
                turf = tile.turf.id
            else:
                turf = tile.turf
            unk = tile.unknown

            if rle > 0:
                if (area, turf, unk) == last_match:
                    rle += 1
                    if rle > 255:
                        raise RuntimeError("RLE overflow.")
                    elif rle == 255:
                        self._uarch(area)
                        self._uarch(turf)
                        self._uarch(unk)
                        self._uint8(255)
                        rle = 0
                else:
                    self._uarch(last_match[0])
                    self._uarch(last_match[1])
                    self._uarch(last_match[2])
                    self._uint8(rle)
                    rle = 1
                    last_match = (area, turf, unk)
            else:
                rle = 1
                last_match = (area, turf, unk)
            length -= 1
            x += 1
            if x >= lxr:
                x = 0
                y += 1
                if y >= lyr:
                    y = 0
                    z += 1
                    if z >= lzr and length > 0:
                        raise ValueError("Length of tile write longer than total length of map.")
        if rle > 0:
            self._uarch(last_match[0])
            self._uarch(last_match[1])
            self._uarch(last_match[2])
            self._uint8(rle)

    def _write_types(self):
        type_count = len(self.dmb.types)
        self._uarch(type_count)
        for t in self.dmb.types:
            self._uarch(t.path)
            self._uarch(t.parent)
            self._uarch(t.name)
            self._uarch(t.desc)
            self._uarch(t.icon)
            self._uarch(t.icon_state)
            self._uint8(t.dir)
            self._uint8(t._unknown1)
            if t._unknown1 == 15:
                self._bytes(t._fdata1)
            self._uarch(t.text)
            self._uarch(t.suffix)
            self._uint16(t.maptext_width)
            self._uint16(t.maptext_height)
            if self.dmb.world.min_client > 507:
                self._uint16(t.maptext_x)
                self._uint16(t.maptext_y)
            self._uarch(t.maptext)
            self._uint32(t.flags)
            self._uarch(t.verb_list)
            self._uarch(t.proc_list)
            self._uarch(t._unknown3)
            self._uarch(t._unknown4)
            self._uarch(t.variable_list)
            self._float32(t.layer)
            if self.dmb.world.min_client >= 500:
                self._uint8(t._unknown2)
                if t._unknown2 > 0:
                    self._bytes(t._fdata4)
            self._uarch(t.builtin_variable_list)

    def _ffwd(self, count):
        return self.writer.seek(count, io.SEEK_CUR)

    def _write_string_length(self, strlen):
        while strlen >= 65535:
            s = (strlen ^ self._ffwd(0)) & 65535
            self._uint16(s)
            strlen -= 65535
        s = (strlen ^ self._ffwd(0)) & 65535
        self._uint16(s)

    def _crc(self, b, curr):
        return byond32(curr, b, null_terminate=True)

    def _write_strings(self):
        strcrc = np.uint32(0xFFFFFFFF)
        string_count = len(self.dmb.strings)
        self._uarch(string_count)
        for s in self.strings:
            strcrc = self._crc(s, strcrc)
            self._write_string_length(len(s))
            s = RawString(s, self._ffwd(0), mode=constants.raw_string_mode_decrypted, lazy=True)
            data = s.encrypt(True)
            self._bytes(data)
        self._uint32(strcrc)

    def _write_mobs(self):
        mob_count = len(self.dmb.mobs)
        self._uarch(mob_count)
        for mob in self.dmb.mobs:
            self._bytes(mob._fdata1)
            self._uint8(mob._unknown)
            if (mob._unknown & 0x80) > 0:
                self._bytes(mob._fdata2)

    def _write_data(self):
        data_count = len(self.dmb.data)
        self._uarch(data_count)
        for data in self.dmb.data:
            dlen = len(data)
            if self.dmb.bit32:
                dlen /= 4
            else:
                dlen /= 2
            self._uint16(int(dlen))
            self._bytes(data)

    def _write_procs(self):
        proc_count = len(self.written_procs)
        self._uarch(proc_count)
        for proc in self.dmb.procs:
            self._uarch(proc.path)
            self._uarch(proc.name)
            self._uarch(proc.desc)
            self._uarch(proc.category)
            self._uint8(proc.range)
            self._uint8(proc.access)
            self._uint8(proc.flags)
            if (proc.flags & 0x80) > 0:
                self._uint32(proc.ext_flags)
                self._uint8(proc.invisibility)
            self._uarch(proc.data)
            self._uarch(proc.variable_list)
            self._uarch(proc.argument_list)

    def _write_vars(self):
        var_count = len(self.written_vars)
        self._uarch(var_count)
        for var in self.dmb.vars:
            self._uint8(var.value._typeid)
            self._uint32(var.value._value)
            self._uarch(var.name)

    def _write_instances(self):
        inst_count = len(self.dmb.instances)
        self._uarch(inst_count)
        for inst in self.dmb.instances:
            self._uint8(inst.value._typeid)
            self._uint32(inst.value._value)
            self._uarch(inst.initializer)  # TODO: unresolve

    def _write_argprocs(self):
        ap_count = len(self.dmb.argprocs)
        self._uarch(ap_count)
        for a in self.dmb.argprocs:
            self._uarch(a)

    def _write_mappops(self):
        total = 0
        tid = 0
        last_tid = 0
        for zlevel in self.dmb.tiles:
            for row in zlevel:
                for tile in row:
                    total += len(tile.instances)
        self._uint32(total)
        for zlevel in self.dmb.tiles:
            for row in zlevel:
                for tile in row:
                    for inst in tile.instances:
                        self._uint16(tid - last_tid)
                        last_tid = tid
                        iid = self.dmb.instances.index(inst)
                        self._uarch(iid)
                    tid += 1

    def _write_resources(self):
        res_count = len(self.dmb.resources)
        self._uarch(res_count)
        for r in self.dmb.resources:
            self._uint32(r.hash)
            self._uint8(r.typeid)

    def _write_extended_data(self):
        self._uarch(self.dmb.world.default_mob)
        self._uarch(self.dmb.world.default_turf)
        self._uarch(self.dmb.world.default_area)
        self._uarch(self.dmb.world.world_procs)
        self._uarch(self.dmb.world.global_init)
        self._uarch(self.dmb.world.world_domain)
        self._uarch(self.dmb.world.world_name)
        self._uint16(self.dmb.world.tick_lag)
        self._uint16(self.dmb.world.unknown1)
        self._uarch(self.dmb.world.client_type)
        self._uarch(self.dmb.world.image_type)
        self._uint8(self.dmb.world.lazy_eye)
        self._uint8(self.dmb.world.client_dir)
        self._uint8(self.dmb.world.control_freak)
        self._uint16(self.dmb.world.unknown2)
        self._uarch(self.dmb.world.client_script)
        self._uint16(self.dmb.world.unknown3)
        self._uint8(self.dmb.world.view_width)
        self._uint8(self.dmb.world.view_height)
        self._uarch(self.dmb.world.hub_password)
        self._uarch(self.dmb.world.world_status)
        self._uint16(self.dmb.world.unknown4)
        self._uint16(self.dmb.world.unknown5)
        self._uint32(self.dmb.world.version)
        self._uint16(self.dmb.world.cache_lifespan)
        self._uarch(self.dmb.world.default_command_text)
        self._uarch(self.dmb.world.default_command_prompt)
        self._uarch(self.dmb.world.hub_path)
        self._uarch(self.dmb.world.unknown6)
        self._uarch(self.dmb.world.unknown7)
        self._uint16(self.dmb.world.icon_width)
        self._uint16(self.dmb.world.icon_height)
        self._uint16(self.dmb.world.map_format)

    def write(self):
        if self.dmb.string_mode == constants.string_mode_strings:
            self.strings = [RawString(s, 0, mode=constants.raw_string_mode_string, lazy=True).encode() for s in self.dmb.strings]
        else:
            self.strings = self.dmb.strings
        self.string_mem_len = 0
        for s in self.strings:
            self.string_mem_len += len(s) + 1

        self._write_version_data()
        self._uint32(self.dmb.flags)
        self._uint16(self.dmb.world.map_x)
        self._uint16(self.dmb.world.map_y)
        self._uint16(self.dmb.world.map_z)
        self._write_tiles()
        self._uint32(self.string_mem_len)

        self._write_types()
        self._write_mobs()

        self._write_strings()
        self._write_data()
        self._write_procs()
        self._write_vars()
        self._write_argprocs()
        self._write_instances()
        self._write_mappops()
        self._write_extended_data()
        self._write_resources()



from .dmb import Tile, Mob, Proc, WorldData, Var, Instance, Resource, Type, RawString
from .dmbwriter import DmbWriter
from .tree import ObjectTree
from . import value, constants
from .crypt import byond32
from blist import *
import numpy as np
import re
import io
import struct
import time


class DmbFileError(ValueError):
    pass


class TileGenerator:
    def __init__(self, reader, count):
        self.reader = reader
        self.len = count
        self.curr = None
        self.rle_count = 0
        self.calls = 0

    def gen(self, count):
        while self.len > 0 and count > 0:
            if self.curr is None or self.rle_count == 0:
                self.curr = (self.reader._uarch(), self.reader._uarch(), self.reader._uarch())
                self.rle_count = self.reader._uint8()
            self.rle_count -= 1
            self.len -= 1
            count -= 1
            self.calls += 1
            yield Tile(self.curr[0], self.curr[1], self.curr[2])


class Dmb:
    def __init__(self, dmbname, throttle=False, verbose=False, string_mode=constants.string_mode_default, check_string_crc=False):
        self.string_mode = string_mode if string_mode != constants.string_mode_default else contants.string_mode_byte_strings
        self.check_string_crc = check_string_crc
        self.reader = open(dmbname, 'rb')
        self.bit32 = False
        self.throttle = throttle
        self.ops = 0
        self.world = WorldData()
        self._parse_version_data()
        if verbose:
            print("Compiled with byond {0} (requires {1} server, {2} client)".format(self.world.world_version, self.world.min_server, self.world.min_client))

        self.flags = self._uint32()
        self.world.map_x, self.world.map_y, self.world.map_z = (self._uint16(), self._uint16(), self._uint16())
        self.bit32 = (self.flags & 0x40000000) > 0

        if verbose:
            print("{0}-bit dmb".format(32 if self.bit32 else 16))

        tilegen = TileGenerator(self, self.world.map_x * self.world.map_y * self.world.map_z)
        self.tiles = [[[tile for tile in tilegen.gen(self.world.map_x)] for j in range(self.world.map_y)] for i in range(self.world.map_z)]
        self._uint32()  # drop this

        self.no_parent_type = Type("/", None)

        type_count = self._uarch()
        if verbose:
            print("{0} types".format(type_count))
        self.types = blist([t for t in self._typegen(type_count)])

        mob_count = self._uarch()
        if verbose:
            print("{0} mobs".format(mob_count))
        self.mobs = blist([m for m in self._mobgen(mob_count)])

        self.strcrc = np.uint32(0)
        string_count = self._uarch()
        if verbose:
            print("{0} strings".format(string_count))
        self.strings = blist([s for s in self._stringgen(string_count)])
        crc = self._uint32()  # CRC
        if self.check_string_crc:
            if crc != self.strcrc:
                raise DmbFileError("String table CRC mismatch (expected: {0}, got: {1})".format(crc, self.strcrc))
            elif verbose:
                print("String table CRC check passed.")

        data_count = self._uarch()
        if verbose:
            print("{0} data".format(data_count))
        self.data = blist([d for d in self._datagen(data_count)])

        proc_count = self._uarch()
        if verbose:
            print("{0} procs".format(proc_count))
        self.procs = blist([p for p in self._procgen(proc_count)])

        var_count = self._uarch()
        if verbose:
            print("{0} vars".format(var_count))
        self.variables = blist([v for v in self._vargen(var_count)])

        argproc_count = self._uarch()
        if verbose:
            print("{0} argprocs".format(argproc_count))
        self.argprocs = blist([v for v in self._argprocgen(argproc_count)])

        instance_count = self._uarch()
        if verbose:
            print("{0} instances".format(instance_count))
        self.instances = blist([i for i in self._instancegen(instance_count)])

        mappop_count = self._uint32()
        if verbose:
            print("{0} mappops".format(mappop_count))
        self._populate_map(mappop_count)

        self._parse_extended_data()

        res_count = self._uarch()
        if verbose:
            print("{0} resources".format(res_count))
        self.resources = blist([r for r in self._resourcegen(res_count)])

        self.reader.close()
        self.reader = None

    def __del__(self):
        if self.reader is not None:
            self.reader.close()

    def _shift_coords(self, move_count, x, y, z):
        x += move_count

        while x >= self.world.map_x:
            x -= self.world.map_x
            y += 1
        while x < 0:
            x += self.world.map_x
            y -= 1

        while y >= self.world.map_y:
            y -= self.world.map_y
            z += 1
        while y < 0:
            y += self.world.map_y
            z -= 1

        return (x, y, z)

    def _populate_map(self, count):
        x = 0
        y = 0
        z = 0
        tile = self.tile(x, y, z)
        for move_count, instanceid in self._mappopgen(count):
            if move_count > 0:
                x, y, z = self._shift_coords(move_count, x, y, z)
                tile = self.tile(x, y, z)
            tile.instances.append(self._resolve_instance(instanceid))

    def _unpack_arch(self, bs):
        return struct.unpack("<" + (("I" if self.bit32 else "H") * int(len(bs) / (4 if self.bit32 else 2))), bs)

    def tile(self, x, y=None, z=None):
        if isinstance(x, tuple):
            z = x[2]
            y = x[1]
            x = x[0]
        if z > len(self.tiles):
            raise IndexError("Z out of map range.")
        if y > len(self.tiles[z]):
            raise IndexError("Y out of map range.")
        if x > len(self.tiles[z][y]):
            raise IndexError("X out of map range.")
        return self.tiles[z][y][x]

    def _uarch(self):
        if self.bit32:
            return self._uint32()
        else:
            return self._uint16()

    def _ffwd(self, seek):
        return self.reader.seek(seek, io.SEEK_CUR)

    def _ffwdarch(self, seek32, seek16):
        seek = seek16
        if self.bit32:
            seek = seek32
        return self._ffwd(seek)

    def _nbytes(self, n):
        return self.reader.read(n)

    def _nbytesarch(self, n32, n16):
        read = n16
        if self.bit32:
            read = n32
        return self._nbytes(read)

    def _float32(self):
        b = self.reader.read(4)
        if len(b) < 4:
            raise EOFError("Read beyond end of file.")
        return struct.unpack('<f', b)[0]    # interprets 4 bytes as little-endian float

    def _uint32(self):
        b = self.reader.read(4)
        if len(b) < 4:
            raise EOFError("Read beyond end of file.")
        return struct.unpack('<I', b)[0]    # interprets 4 bytes as little-endian unsigned long

    def _uint16(self):
        b = self.reader.read(2)
        if len(b) < 2:
            raise EOFError("Read beyond end of file.")
        return struct.unpack('<H', b)[0]    # interprets 2 bytes as little-endian unsigned short

    def _uint8(self):
        b = self.reader.read(1)
        if len(b) < 1:
            raise EOFError("Read beyond end of file.")
        return struct.unpack('<B', b)[0]    # interprets 1 byte as an unsigned byte

    def _byte(self):
        return self.reader.read(1)

    def _parse_version_data(self):
        ver_str = self._read_bytes_until(b'\x0A').decode('ascii')
        match = re.search(r'world bin v([0-9]*)', ver_str)
        if match is not None:
            self.world.world_version = int(match.group(1))
        else:
            raise DmbFileError('Cannot parse world version from dmb file.')
        ver_str = self._read_bytes_until(b'\x0A').decode('ascii')
        match = re.search(r'min compatibility v([0-9]*) ([0-9]*)', ver_str)
        if match is not None:
            self.world.min_server = int(match.group(1))
            self.world.min_client = int(match.group(2))
        else:
            raise DmbFileError('Cannot parse compatibility version from dmb file.')

    def _parse_extended_data(self):
        self.world.default_mob = self._uarch()
        self.world.default_turf = self._uarch()
        self.world.default_area = self._uarch()
        self.world.world_procs = self._uarch()
        self.world.global_init = self._uarch()
        self.world.world_domain = self._uarch()
        self.world.world_name = self._uarch()
        self.world.tick_lag = self._uint16()
        self.world.unknown1 = self._uint16()
        self.world.client_type = self._uarch()
        self.world.image_type = self._uarch()
        self.world.lazy_eye = self._uint8()
        self.world.client_dir = self._uint8()
        self.world.control_freak = self._uint8()
        self.world.unknown2 = self._uint16()
        self.world.client_script = self._uarch()
        self.world.unknown3 = self._uint16()
        self.world.view_width = self._uint8()
        self.world.view_height = self._uint8()
        self.world.hub_password = self._uarch()
        self.world.world_status = self._uarch()
        self.world.unknown4 = self._uint16()
        self.world.unknown5 = self._uint16()
        self.world.version = self._uint32()
        self.world.cache_lifespan = self._uint16()
        self.world.default_command_text = self._uarch()
        self.world.default_command_prompt = self._uarch()
        self.world.hub_path = self._uarch()
        self.world.unknown6 = self._uarch()
        self.world.unknown7 = self._uarch()
        self.world.icon_width = self._uint16()
        self.world.icon_height = self._uint16()
        self.world.map_format = self._uint16()

    def write(self, dmbname):
        writer = DmbWriter(dmbname, self)
        writer.write()

    def _throttle(self):
        if self.throttle:
            self.ops += 1
            if self.ops >= 1000:
                time.sleep(0.5)
                self.ops = 0

    def _typegen(self, count):
        typeid = 0
        while count > 0:
            path = self._uarch()
            parent = self._uarch()
            curr = Type(path, parent)

            curr.id = typeid
            typeid += 1

            curr.name = self._uarch()
            curr.desc = self._uarch()
            curr.icon = self._uarch()
            curr.icon_state = self._uarch()
            curr.dir = self._uint8()

            curr._unknown1 = self._uint8()
            if curr._unknown1 == 15:
                curr._fdata1 = self._nbytes(4)
            curr.text = self._uarch()
            curr.suffix = self._uarch()  # suffix
            curr.maptext_width = self._uint16()
            curr.maptext_height = self._uint16()
            if self.world.min_client > 507:
                curr.maptext_x = self._uint16()
                curr.maptext_y = self._uint16()
            curr.maptext = self._uarch()
            curr.flags = self._uint32()
            curr.verb_list = self._uarch()
            curr.proc_list = self._uarch()
            curr._unknown3 = self._uarch()
            curr._unknown4 = self._uarch()
            curr.variable_list = self._uarch()
            curr.layer = self._float32()
            if self.world.min_client >= 500:
                curr._unknown2 = self._uint8()
                if curr._unknown2 > 0:
                    curr._fdata4 = self._nbytes(24)
            curr.builtin_variable_list = self._uarch()
            count -= 1
            yield curr
            self._throttle()

    def _mobgen(self, count):
        while count > 0:
            mob = Mob()
            mob._fdata1 = self._nbytesarch(8, 4)
            mob._unknown = self._uint8()
            if (mob._unknown & 0x80) > 0:
                mob._fdata2 = self._nbytes(6)
            count -= 1
            yield mob
            self._throttle()

    def _crc(self, b):
        self.strcrc = byond32(self.strcrc, b, null_terminate=True)

    def _stringgen(self, count):
        while count > 0:
            c = self.reader.seek(0, io.SEEK_CUR)
            lb = self._uint16()
            strlen = (lb ^ c) & 65535
            lastread = strlen
            while strlen == 65535:
                c += 2
                lastread = self._uint16()
                nextd = (lastread ^ c) & 65535
                strlen += nextd
            key = c + 2
            count -= 1
            estr = self._nbytes(strlen)
            string = RawString(bytearray(estr), key, lazy=True)
            if self.check_string_crc:
                self._crc(string.decrypt())
            if self.string_mode == constants.string_mode_byte_strings:
                yield string.decrypt()
            elif self.string_mode == constants.string_mode_strings:
                yield string.decode()
            self._throttle()

    def _datagen(self, count):
        while count > 0:
            mult = 2
            if self.bit32:
                mult = 4
            datalen = self._uint16() * mult
            data = self._nbytes(datalen)
            count -= 1
            yield data
            self._throttle()

    def _procgen(self, count):
        while count > 0:
            ret = Proc()
            ret.path = self._uarch()
            ret.name = self._uarch()
            ret.desc = self._uarch()
            ret.category = self._uarch()
            ret.range = self._uint8()
            ret.access = self._uint8()
            ret.flags = self._uint8()
            if ret.flags & 0x80 > 0:
                ret.ext_flags = self._uint32()
                ret.invisibility = self._uint8()
            ret.data = self._uarch()
            ret.variable_list = self._uarch()
            ret.argument_list = self._uarch()
            count -= 1
            yield ret
            self._throttle()

    def _vargen(self, count):
        while count > 0:
            ret = Var()
            typeid = self._uint8()
            typeval = self._uint32()
            ret.name = self._uarch()
            ret.value = value.decode(typeid, typeval)
            count -= 1
            yield ret
            self._throttle()

    def _instancegen(self, count):
        while count > 0:
            ret = Instance()
            typeid = self._uint8()
            typeval = self._uint32()
            ret.initializer = self._uarch()
            ret.value = value.decode(typeid, typeval)
            count -= 1
            yield ret
            self._throttle()

    def _argprocgen(self, count):
        while count > 0:
            count -= 1
            yield self._uarch()
            self._throttle()

    def _mappopgen(self, count):
        while count > 0:
            move = self._uint16()
            instance = self._uarch()
            count -= 1
            yield (move, instance)
            self._throttle()

    def _resourcegen(self, count):
        while count > 0:
            rhash = self._uint32()
            typeid = self._uint8()
            count -= 1
            yield Resource(typeid, rhash)
            self._throttle()

    def _bytegen(self, delimiter):
        eof = False
        while not eof:
            byte = self.reader.read(1)
            if byte == b'':
                eof = True
                break
            if byte == delimiter:
                break
            yield byte
            self._throttle()

    def _read_bytes_until(self, delimiter):
        return b''.join([b for b in self._bytegen(delimiter)])

    def _resolve_data(self, dataid):
        if dataid == 65535:
            return None
        try:
            return self.data[dataid]
        except:
            print(dataid)
            raise

    def _resolve_resource(self, resourceid):
        if resourceid == 65535:
            return None
        try:
            return self.resources[resourceid]
        except:
            print(resourceid)
            raise

    def _resolve_proc(self, procid):
        try:
            return self.procs[procid]
        except:
            print(procid)
            raise

    def _resolve_var(self, varid):
        try:
            return self.variables[varid]
        except:
            print(varid)
            raise

    def _resolve_instance(self, instanceid):
        try:
            return self.instances[instanceid]
        except:
            print(instanceid)
            raise

    def _resolve_string(self, stringid, mode=constants.string_mode_default):
        if stringid == 0xFFFF:
            return None
        ret = self.strings[stringid]
        if isinstance(ret, RawString):
            val = ret.decrypt()
            self.strings[stringid] = val
            ret = val
        dcmode = self.string_mode if mode == constants.string_mode_default else mode
        if dcmode != self.string_mode:
            if dcmode == constants.string_mode_strings:
                return RawString(ret, 0, mode=constants.raw_string_mode_decrypted).decode()
            elif dcmode == constants.string_mode_byte_strings:
                return RawString(ret, 0, mode=constants.raw_string_mode_string).encode()
        return ret



#from setuptools import setup, find_packages
from setuptools import *

description='DMB file parser for python'

setup(
    name='pydmb',
    version='0.2.0',
    description=description,
    long_description=description,
    url='https://github.com/baliame/pydmb',
    author='Baliame',
    author_email='akos.toth@cheppers.com',
    license='MIT',
    classifiers=[
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    keywords='byond dmb development library',
    install_requires=[],
    packages=find_packages(),
)



from . import constants
import numpy as np


n_8 = np.uint8(8)
n_24 = np.uint8(24)
n_255 = np.uint8(255)


def byond32(initial, data, null_terminate=False):
    crc = initial
    for byte in data:
        b = np.uint8(byte)
        tab = constants.byond32_tab[(b ^ (crc >> n_24)) & n_255]
        crc = (tab ^ (crc << n_8))
    if null_terminate:
        b = np.uint8(0)
        tab = constants.byond32_tab[(b ^ (crc >> n_24)) & n_255]
        crc = (tab ^ (crc << n_8))
    return crc
